/* Linear Program Graph Clustering
 *
 * Copyright (C) 2008 -- Sebastian Nowozin <nowozin@gmail.com>
 */

// References
//
// [Deza1992], "Clique-web facets for multicut polytopes",
//    M. Deza, M. Grötschel, M. Laurent, 1992.
// [Jansen1997], "Sensitivity Analysis in Linear Programming: just be careful!"
//    B. Jansen, J.J. de Jong, C. Roos, T. Terlaky

#include <algorithm>
#include <iostream>
#include <functional>
#include <ext/hash_set>
#include <ext/hash_map>
#include <ext/numeric>
#include <numeric>
#include <limits>
#include <list>
#include <boost/progress.hpp>
#include <boost/timer.hpp>

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <coin/OsiSolverParameters.hpp>

#include "GraphClustering.h"

namespace Clustering {

GraphClustering::GraphClustering(bool do_separate_oddwheel, int max_iter)
	: si(NULL), dynamic_constraint_start(0),
	do_separate_oddwheel(do_separate_oddwheel), max_iter(max_iter),
	doing_stability_analysis(false) {
}

GraphClustering::~GraphClustering() {
	if (si != NULL)
		delete si;
}

void GraphClustering::Initialize(const gmm::row_matrix<gmm::wsvector<double> >& K) {
	// Fill edgemap, nodemap and edgevar from K.
	unsigned int ei = 0;
	nodemap.resize(gmm::mat_nrows(K));
	edgemap.clear();
	obj.clear();
	for (unsigned int row = 0; row < gmm::mat_nrows(K); ++row) {
		const gmm::simple_vector_ref<const gmm::wsvector<double>*>
			row_vector = gmm::mat_const_row(K, row);
		gmm::linalg_traits<gmm::wsvector<double> >::const_iterator it1 =
			gmm::vect_const_begin(row_vector);
		gmm::linalg_traits<gmm::wsvector<double> >::const_iterator ite1 =
			gmm::vect_const_end(row_vector);

		for (; it1 != ite1; ++it1) {
			// Skip tril elements (row > col)
			if (row >= it1.index())
				continue;

			// Add edge
			edgemap.push_back(
				std::pair<unsigned int, unsigned int>(row, it1.index()));
			obj.push_back(*it1);

			// Add edge reference to both nodes
			nodemap[row].insert(ei);
			nodemap[it1.index()].insert(ei);

			ei += 1;
		}
	}
	edgevar.resize(edgemap.size());
#if 0
	if (minimum_cluster_size > 0 || maximum_cluster_size > 0) {
		if (edgemap.size() != (gmm::mat_nrows(K) * (gmm::mat_nrows(K) - 1))/2) {
			std::cerr << "ERROR: The minimum/maximum cluster size constraints "
				<< "are only legal for completely connected graphs." << std::endl;
			assert(0);
			return;
		}
	}
#endif
	InternalInitialize();
}

void GraphClustering::Initialize(const gmm::dense_matrix<double>& K) {
	// Fill edgemap, nodemap and edgevar from K.
	unsigned int ei = 0;
	nodemap.resize(gmm::mat_nrows(K));
	edgemap.clear();
	obj.clear();
	assert(gmm::mat_nrows(K) == gmm::mat_ncols(K));
	for (unsigned int row = 0; row < gmm::mat_nrows(K); ++row) {
		for (unsigned int col = (row + 1); col < gmm::mat_ncols(K); ++col) {
			// Add edge
			edgemap.push_back(
				std::pair<unsigned int, unsigned int>(row, col));
			obj.push_back(K(row, col));

			// Add edge reference to both nodes
			nodemap[row].insert(ei);
			nodemap[col].insert(ei);
			ei += 1;
		}
	}
	edgevar.resize(edgemap.size());

	InternalInitialize();
}

void GraphClustering::InternalInitialize() {
	// Setup LP
	si = new OsiClpSolverInterface;

	// Row-ordered constraint matrix (appending rows will be cheaper than
	// appending columns)
	CoinPackedMatrix* matrix = new CoinPackedMatrix(false, 0, 0);
	matrix->setDimensions(0, edgevar.size());

	double* rowLB = NULL;
	double* rowUB = NULL;
#if 0
	if (minimum_cluster_size > 0 || maximum_cluster_size > 0) {
		unsigned int sample_count = gmm::mat_nrows(K);
		rowLB = static_cast<double*>(calloc(sizeof(double), sample_count));
		rowUB = static_cast<double*>(calloc(sizeof(double), sample_count));
		for (unsigned int sample = 0; sample < sample_count; ++sample) {
			// Set coefficients for all adjacent edges
			CoinPackedVector row;
			const std::set<unsigned int>& edges_to_node = nodemap[sample];
			for (std::set<unsigned int>::const_iterator ei = edges_to_node.begin();
				ei != edges_to_node.end(); ++ei)
					row.insert(*ei, 1.0);

			// Bounds
			rowLB[sample] = (maximum_cluster_size > 0) ?
				(sample_count - 1 - maximum_cluster_size) : 0;
			rowUB[sample] = (minimum_cluster_size > 0) ?
				(sample_count - 1 - minimum_cluster_size) : sample_count;
			matrix->appendRow(row);
		}
		dynamic_constraint_start = sample_count;
	} else
#endif
	dynamic_constraint_start = 0;

	/* Variables lower/upper bounds all [0;1]
	 */
	double* varLB = static_cast<double*>(calloc(sizeof(double), edgevar.size()));
	double* varUB = static_cast<double*>(calloc(sizeof(double), edgevar.size()));
	for (unsigned int n = 0; n < edgevar.size (); ++n) {
		varLB[n] = 0.0;
		varUB[n] = 1.0;
	}

	// FIXME: maybe new double[...]
	double* objective = static_cast<double*>(calloc(sizeof(double), obj.size()));
	std::copy(obj.begin(), obj.end(), objective);

	si->loadProblem(*matrix, varLB, varUB, objective, rowLB, rowUB);
	si->setObjSense(1);

	free(objective);
	free(varLB);
	free(varUB);
	if (rowLB != NULL)
		free(rowLB);
	if (rowUB != NULL)
		free(rowUB);
	delete matrix;

	// Use interior-point
	bool interior_point = false;
	if (interior_point) {
		ClpSolve lp_options;
		// Don't do crossover to a feasible basis.
		lp_options.setSolveType(ClpSolve::useBarrierNoCross);
		lp_options.setPresolveType(ClpSolve::presolveOn);
		//lp_options.setPresolveType(ClpSolve::presolveOff);
		lp_options.setSpecialOption(6, 1);
		si->setSolveOptions(lp_options);
	}
}

void GraphClustering::Solve(double theta_left, double theta_right) {
	theta_path.clear();
	theta_edgevars.clear();
	theta_iters_exceeded.clear();

	// Save original objective once
	std::vector<double> obj_base = obj;
	std::fill(edgevar.begin(), edgevar.end(), 0.0);

	// Handle special case of theta_right = infinity, by setting it such that
	// all objective coefficients become positive
	double theta_cur = theta_right;
	if (isinf(theta_cur) == 1) {
		double obj_min = *std::min_element(obj.begin(), obj.end());
		if (obj_min >= 0.0)
			theta_cur = 0.0;	// trivial solution
		else
			theta_cur = -obj_min + 1e-5;	// make all elements positive
	}

	// Handle theta_left = -infinity
	if (isinf(theta_left) == -1) {
		double obj_max = *std::max_element(obj.begin(), obj.end());
		if (obj_max <= 0.0)
			theta_left = 0.0;
		else
			theta_left = -obj_max - 1e-5;	// make all elements negative
	}

	double* obj_arr = new double[obj.size()];

	// d = 1, shifting all edge weights positively
	std::vector<double> d_ones(obj.size());
	std::fill(d_ones.begin(), d_ones.end(), 1.0);

	std::vector<double> sol_prev;
	double last_stable = theta_right;
	double cur_stable = theta_right;
	std::vector<double> theta_edgevar;
	unsigned int iters = 0;
	bool no_path = fabs(theta_right - theta_left) < 1e-8;
	while (theta_cur >= theta_left) {
		// Set the objective to the current shift minus a small number
		std::transform(obj_base.begin(), obj_base.end(),
			obj_arr, std::bind2nd(std::plus<double>(), theta_cur));
		si->setObjective(obj_arr);

		// First iteration if theta_right is infinity
		std::cout << "FIXED with theta = " << theta_cur << std::endl;
		iters = SolveFixed();
		std::cout << "      FIXED obj " << Objective()
			<< ", took " << (iters == 0 ? max_iter : iters)
			<< " iterations" << std::endl;

		// Track solution
		const std::vector<double>& sol = Solution();
		double diff = -std::numeric_limits<double>::infinity();
		double diff_1 = 0.0;
		if (sol_prev.size() > 0) {
			diff = -std::numeric_limits<double>::infinity();
			diff_1 = 0.0;
			// Calculate max norm between this and previous solution
			for (unsigned int n = 0; n < sol.size(); ++n) {
				diff_1 += fabs(sol[n] - sol_prev[n]);
				if (fabs(sol[n] - sol_prev[n]) > diff)
					diff = fabs(sol[n] - sol_prev[n]);
			}
			std::cout << "      FIXED max norm diff " << diff << std::endl;
			std::cout << "      FIXED 1-norm diff " << diff_1 << std::endl;

			// If the solution changed, record next point in solution path and
			// create new theta window.  If the solution did not change,
			// we have nothing to do, and keep tracking
			if (diff > 1e-3) {
				// Past solution is stable between [cur_stable ; last_stable]
				theta_path.push_back(std::pair<double, double>(cur_stable, last_stable));
				theta_edgevars.push_back(theta_edgevar);
				theta_iters_exceeded.push_back(iters == 0);

				// Next stability range starts at the end of this one
				last_stable = cur_stable;
				theta_edgevar = edgevar;
			}
		} else {
			std::cout << "      FIXED first solution" << std::endl;
			theta_edgevar = edgevar;
		}
		sol_prev = sol;
		if (no_path)
			break;

		// Compute guaranteed stability range for "\theta 1"-direction,
		// applying a constant shift to all weights
		double stable_theta_left;
		double stable_theta_right;
		//ComputeStabilityRange(d_ones, stable_theta_left, stable_theta_right);
		ComputeStabilityRangeLP(d_ones, stable_theta_left, stable_theta_right);
		std::cout << "      FIXED stability range [" << stable_theta_left
			<< " to " << stable_theta_right << "]" << std::endl;

		// Update theta by shifting left (towards many clusters)
		assert(stable_theta_left <= 0.0);
		theta_cur += stable_theta_left;
		cur_stable = theta_cur;
		theta_cur -= 1e-5;

		// If the user calls us with [0 ; 0] we should still return the proper
		// stability range.
		if (theta_path.empty() && stable_theta_right > last_stable)
			last_stable = stable_theta_right;
	}
	delete[] obj_arr;

	// Store last solution is stable between [cur_stable ; last_stable]
	theta_path.push_back(std::pair<double, double>(cur_stable, last_stable));
	theta_edgevars.push_back(theta_edgevar);
	theta_iters_exceeded.push_back(iters == 0);

	// Restore objective coefficients
	obj = obj_base;

	for (unsigned int n = 0; n < theta_path.size(); ++n) {
		std::cout << "path solution for [" << theta_path[n].first
			<< " ; " << theta_path[n].second << "]" << std::endl;
	}
}

unsigned int GraphClustering::SolveFixed() {
	// Main loop,
	bool converged = false;
	std::vector<double> objectives;
	std::vector<double> edgevar_previous;
	double obj_last = -std::numeric_limits<double>::infinity();
	unsigned int iter;
	for (iter = 1; converged == false
		&& (max_iter <= 0 || iter <= static_cast<unsigned int>(max_iter));
		++iter) {
		std::cout << "[gclustering] iter " << iter
			<< ", " << si->getNumRows() << " constraints" << std::endl;
		edgevar_previous = edgevar;

		// 1. Solve LP
		si->messageHandler()->setLogLevel(0);	// no verbosity
		boost::timer lp_timer;
		si->resolve();	// Warm-start solving (we only add constraints)
		std::cout << lp_timer.elapsed() << "s in LP solver" << std::endl;

		if (si->isProvenOptimal() == false) {
			std::cerr << "Linear Program Solver failed." << std::endl;
			std::cerr << "Problem: " << si->getNumCols() << " variables, "
				<< si->getNumRows() << " rows." << std::endl;
			si->writeMps ("LP-CRASH", "mps", si->getObjSense());
			std::cerr << "Written problem to file \"LP-CRASH.MPS\" for analysis."
				<< std::endl;
			std::cerr << "         problem sense: "
				<< si->getObjSense() << std::endl;
			std::cerr << "STATUS:  numerical difficulties: "
				<< (si->isAbandoned() ? "YES" : "no") << std::endl;
			std::cerr << "STATUS:       primal infeasible: "
				<< (si->isProvenPrimalInfeasible() ? "YES" : "no") << std::endl;
			std::cerr << "STATUS:         dual infeasible: "
				<< (si->isProvenDualInfeasible() ? "YES" : "no") << std::endl;
			std::cerr << "STATUS: iteration limit reached: "
				<< (si->isIterationLimitReached() ? "YES" : "no") << std::endl;
			exit(EXIT_FAILURE);
		}

		/* Output objectives achieved so far
		 */
		objectives.push_back(si->getObjValue());
		std::cout << "   objs: " << gmm::row_vector(objectives)
			<< std::endl;

		// copy primal solution, which may be infeasible
		unsigned int cols = si->getNumCols();
		std::cout << "   cols: " << cols << std::endl;
		const double* primal = si->getColSolution();
		double max_change = -std::numeric_limits<double>::infinity();
		for (unsigned int ei = 0; ei < edgevar.size(); ++ei) {
			edgevar[ei] = primal[ei];
			if (edgevar[ei] < 0.0)
				edgevar[ei] = 0.0;
#if 0
			std::cout << "edgevar " << ei << " (" << edgemap[ei].first
				<< "-" << edgemap[ei].second << "): " << edgevar[ei]
				<< ", prev " << edgevar_previous[ei]
				<< ", coeff " << obj[ei] << std::endl;
#endif
			double ei_change = fabs(edgevar[ei] - edgevar_previous[ei]);
			if (ei_change > max_change)
				max_change = ei_change;	// ||prev-cur||_\infty
		}
#if 0
		std::cout << "Change max-norm: " << max_change << std::endl;
#endif

		// 2. Remove all inactive constraints
		//    Only do this in case the objective has changed.  Otherwise, we
		// might be at risk of cycling.  To see this effect, see testcase
		// "Leaf9" where this happens.
		double obj_improved = si->getObjValue() - obj_last;
		obj_last = si->getObjValue();
		if (obj_improved > 1e-5) {
			RemoveInactiveConstraints();
		} else {
			std::cout << "Not removing inactive constraints, "
				<< "avoiding possible cycles (progress "
				<< obj_improved << ")." << std::endl;
		}

		// 3. Find and add violated constraints
		converged = (AddViolatedConstraints(si, edgevar) == false);
	}
	unsigned int final_removed = RemoveInactiveConstraints();
	std::cout << "Final pass removed " << final_removed << " constraints."
		<< std::endl;

	if (max_iter <= 0)
		return (iter);

	return (iter > static_cast<unsigned int>(max_iter) ? 0 : iter);
}

unsigned int GraphClustering::RemoveInactiveConstraints() {
	int rows = si->getNumRows();
	std::vector<int> rows_to_delete;
	const double* dual = si->getRowPrice();

	for (int row = dynamic_constraint_start; row < rows; ++row) {
#if 0
		std::cout << "duals, row " << row << " is " << dual[row] << std::endl;
#endif
		if (fabs(dual[row]) < 1e-10)
			rows_to_delete.push_back(row);
	}
	if (rows_to_delete.size() > 0) {
		int* rows_to_delete_arr = new int[rows_to_delete.size()];
		std::copy(rows_to_delete.begin(), rows_to_delete.end(),
			rows_to_delete_arr);
		// because we only remove inactive constraints, the basis should stay
		// the same and warmstart should be possible (see documentation of
		// OsiSolverInterface::deleteRows)
		si->deleteRows(rows_to_delete.size(), rows_to_delete_arr);
		delete[](rows_to_delete_arr);
		std::cout << "    removed " << rows_to_delete.size()
			<< " inactive constraints, "
			<< si->getNumRows() << " active constraints left."
			<< std::endl;
	}
	return (rows_to_delete.size());
}

double GraphClustering::Objective() const {
	double res = 0.0;
	for (unsigned int n = 0; n < obj.size(); ++n)
		res += obj[n] * edgevar[n];
	return (res);
}

const std::vector<double>& GraphClustering::ObjectiveCoefficients() const {
	return (obj);
}

const std::vector<double>& GraphClustering::Solution() const {
	return (edgevar);
}

const std::vector<std::pair<double, double> >&
GraphClustering::ThetaPath() const {
	return (theta_path);
}

const std::vector<std::vector<double> >&
GraphClustering::ThetaSolutions() const {
	return (theta_edgevars);
}

const std::vector<bool>& GraphClustering::ThetaItersExceeded() const {
	return (theta_iters_exceeded);
}

bool GraphClustering::IsFractionalSolution() const {
	return (IsFractionalSolution(edgevar));
}

bool GraphClustering::IsFractionalSolution(const std::vector<double>& ev) const {
	for (unsigned int n = 0; n < ev.size(); ++n) {
		if (ev[n] >= zero_one_tolerance &&
			ev[n] <= (1.0 - zero_one_tolerance))
			return (true);
	}
	return (false);
}

const std::vector<std::pair<unsigned int, unsigned int> >&
GraphClustering::Edgemap() const {
	return (edgemap);
}

void GraphClustering::ComputeStabilityRange(const std::vector<double>& d,
	double& theta_min, double& theta_max) {
	std::vector<double> box_min;
	std::vector<double> box_max;
	ComputeStabilityBox(d, box_min, box_max);

	theta_min = *std::max_element(box_min.begin(), box_min.end());
	theta_max = *std::min_element(box_max.begin(), box_max.end());
}

void GraphClustering::ComputeStabilityRangeLP(const std::vector<double>& d,
	double& theta_min, double& theta_max) {
	// Get key information about current optimal solution
	//double wzopt = Objective();
	double wzopt = si->getObjValue();	// XXX: Objective() is wrong
	const std::vector<double>& zopt = Solution();
	double dzopt = std::inner_product(zopt.begin(), zopt.end(), d.begin(), 0.0);

	// Duplicate model
	OsiClpSolverInterface* si2 = new OsiClpSolverInterface(*si);

	// Setup problem:
	//
	// min_{z,\alpha}  w'z + (w'z^*) \alpha
	// sb.t.           Az <= 0
	//                 - d'z - (d'z^*) \alpha = +/- 1 : \gamma
	//                 0 <= z_m <= -\alpha, for all m

	//   i) remove upper bounds on z_m
	for (unsigned int ei = 0; ei < edgevar.size(); ++ei)
		si2->setColUpper(ei, si2->getInfinity());

	//  ii) add free \alpha variable
	CoinPackedVector col_alpha;
	si2->addCol(col_alpha, -si2->getInfinity(), si2->getInfinity(), wzopt);
	unsigned int alpha_index = edgevar.size();

	// iii) add - d'z - (d'z^*) \alpha = 1 constraint
	CoinPackedVector row_dz;
	for (unsigned int ei = 0; ei < edgevar.size(); ++ei)
		row_dz.insert(ei, -d[ei]);
	row_dz.insert(alpha_index, -dzopt);
	unsigned int gamma_row_index = si2->getNumRows();
	si2->addRow(row_dz, 1.0, 1.0);

	//  iv) add z_m <= -\alpha, for all m
	for (unsigned int ei = 0; ei < edgevar.size(); ++ei) {
		CoinPackedVector row_vub_alpha;
		row_vub_alpha.insert(ei, 1.0);
		row_vub_alpha.insert(alpha_index, 1.0);
		si2->addRow(row_vub_alpha, -si2->getInfinity(), 0.0);
	}

	std::vector<double> edgevar2(edgevar.size());

	int gamma_max_inf = ComputeStabilityRangeLPSingle(si2,
		edgevar2, alpha_index);
	if (gamma_max_inf == 1) {
		theta_max = std::numeric_limits<double>::infinity();
	} else if (gamma_max_inf == -1) {
		// -1: dual infeasible, primal unbounded or infeasible
		theta_max = std::numeric_limits<double>::infinity();
		std::cerr << "         problem sense: "
			<< si2->getObjSense() << std::endl;
		std::cerr << "STATUS:  numerical difficulties: "
			<< (si2->isAbandoned() ? "YES" : "no") << std::endl;
		std::cerr << "STATUS:       primal infeasible: "
			<< (si2->isProvenPrimalInfeasible() ? "YES" : "no") << std::endl;
		std::cerr << "STATUS:         dual infeasible: "
			<< (si2->isProvenDualInfeasible() ? "YES" : "no") << std::endl;
		std::cerr << "STATUS: iteration limit reached: "
			<< (si2->isIterationLimitReached() ? "YES" : "no") << std::endl;
	} else {
		const double* duals = si2->getRowPrice();
		theta_max = duals[gamma_row_index];
	}
	std::cout << "theta_max = " << theta_max << std::endl;

	// Change dual objective coefficient to -1
	si2->setRowBounds(gamma_row_index, -1.0, -1.0);
//	si2->setObjSense(-1.0);

	int gamma_min_inf = ComputeStabilityRangeLPSingle(si2,
		edgevar2, alpha_index);
	if (gamma_min_inf == 1) {
		theta_min = -std::numeric_limits<double>::infinity();
	} else if (gamma_min_inf == -1) {
		// -1: dual infeasible, primal unbounded or infeasible
		theta_min = -std::numeric_limits<double>::infinity();
		std::cerr << "         problem sense: "
			<< si2->getObjSense() << std::endl;
		std::cerr << "STATUS:  numerical difficulties: "
			<< (si2->isAbandoned() ? "YES" : "no") << std::endl;
		std::cerr << "STATUS:       primal infeasible: "
			<< (si2->isProvenPrimalInfeasible() ? "YES" : "no") << std::endl;
		std::cerr << "STATUS:         dual infeasible: "
			<< (si2->isProvenDualInfeasible() ? "YES" : "no") << std::endl;
		std::cerr << "STATUS: iteration limit reached: "
			<< (si2->isIterationLimitReached() ? "YES" : "no") << std::endl;
	} else {
		const double* duals = si2->getRowPrice();
		theta_min = duals[gamma_row_index];
	}
	std::cout << "theta_min = " << theta_min
		<< ", theta_max = " << theta_max << std::endl;
	std::cout << "dzopt = " << dzopt << std::endl;
	delete si2;
}

int GraphClustering::ComputeStabilityRangeLPSingle(OsiClpSolverInterface* si2,
	std::vector<double>& edgevar2, unsigned int alpha_index) {
	bool converged = false;
	while (converged == false) {
		std::cout << "SINGLE ITER" << std::endl;
		si2->messageHandler()->setLogLevel(0);	// no verbosity
		si2->resolve();	// Warm-start solving (we only add constraints)

		if (si2->isProvenOptimal() == false && si2->isProvenPrimalInfeasible())
			return (1);
		if (si2->isProvenOptimal() == false && si2->isProvenDualInfeasible())
			return (-1);

		if (si2->isProvenOptimal() == false) {
			std::cerr << "Linear Program Solver failed." << std::endl;
			std::cerr << "Problem: " << si2->getNumCols() << " variables, "
				<< si2->getNumRows() << " rows." << std::endl;
			si2->writeMps ("LP-CRASH-SENS", "mps", si2->getObjSense());
			std::cerr << "Written problem to file \"LP-CRASH-SENS.MPS\" for analysis."
				<< std::endl;
			std::cerr << "         problem sense: "
				<< si2->getObjSense() << std::endl;
			std::cerr << "STATUS:  numerical difficulties: "
				<< (si2->isAbandoned() ? "YES" : "no") << std::endl;
			std::cerr << "STATUS:       primal infeasible: "
				<< (si2->isProvenPrimalInfeasible() ? "YES" : "no") << std::endl;
			std::cerr << "STATUS:         dual infeasible: "
				<< (si2->isProvenDualInfeasible() ? "YES" : "no") << std::endl;
			std::cerr << "STATUS: iteration limit reached: "
				<< (si2->isIterationLimitReached() ? "YES" : "no") << std::endl;
			exit(EXIT_FAILURE);
		}

		/* Output objectives achieved so far
		 */
		std::cout << "   objective: " << si2->getObjValue() << std::endl;;

		// copy primal solution, which may be infeasible
		const double* primal = si2->getColSolution();
		for (unsigned int ei = 0; ei < edgevar2.size(); ++ei) {
			edgevar2[ei] = primal[ei];
			if (edgevar2[ei] < 0.0)
				edgevar2[ei] = 0.0;
		}
		std::cout << "edgevar: " << gmm::row_vector(edgevar2) << std::endl;

		double alpha = primal[alpha_index];
		std::cout << "   alpha: " << alpha << std::endl;

		// 3. Find and add violated constraints
#if 1
		// Rescale z by -(1/alpha)
		assert(alpha < 0.0);
		for (unsigned int ei = 0; ei < edgevar2.size(); ++ei)
			edgevar2[ei] *= -(1.0 / alpha);
#endif
		// XXX: hack for rhs of odd-wheels
		doing_stability_analysis = true;
		converged = (AddViolatedConstraints(si2, edgevar2) == false);
		doing_stability_analysis = false;
	}
	return (0);
}

// Most general form sensitivity analysis.  Given a currently optimal
// solution, and a cost direction d, compute for each edge cost an individual
// stability range for which the current solution/basis remains optimal.
// Note that the for each variable e_j, the stability range is of the form
// [-infinity; a] with a >= 0, or [a; infinity] with a <= 0, or [0; 0].
void GraphClustering::ComputeStabilityBox(const std::vector<double>& d,
	std::vector<double>& theta_min, std::vector<double>& theta_max) {
	assert(d.size() == edgevar.size());
	theta_min.resize(d.size());
	std::fill(theta_min.begin(), theta_min.end(),
		-std::numeric_limits<double>::infinity());	// default: -\infty
	theta_max.resize(d.size());
	std::fill(theta_max.begin(), theta_max.end(),	// default: \infty
		std::numeric_limits<double>::infinity());

	// Limit case of trivial solutions does need special care as reduced costs
	// are not available (no constraints).
	// Then, the condition is \theta d >= -w, thus if theta_i >= -w_i / d_i,
	// then the basis remains optimal.
	if (si->getNumRows() == 0) {
		for (unsigned int n = 0; n < d.size(); ++n) {
			double denum = d[n];
			double num = obj[n];
			double stability_theta = 0.0;
			// Three cases, as below
			if (fabs(num) > 1e-7 && fabs(denum) > 1e-7) {
				stability_theta = - num / denum;
				if (stability_theta >= 0.0) {
					theta_max[n] = stability_theta;
				} else {
					theta_min[n] = stability_theta;
				}
			} else if (fabs(denum) > 1e-7) {
				theta_min[n] = theta_max[n] = 0.0;
			} else {
			}
		}
		return;
	}

	// \theta r(d) >= - r(w), where r(c) is the reduced cost for the
	// cost vector c.
	si->enableSimplexInterface(true);
	assert(si->canDoSimplexInterface() > 0);
	si->enableFactorization();
	const double* red_cost_w_arr = si->getReducedCost();
	assert(red_cost_w_arr != NULL);
	std::vector<double> red_cost_w(si->getNumCols());
	assert(si->getNumCols() > 0);
	std::copy(red_cost_w_arr, red_cost_w_arr + si->getNumCols(), red_cost_w.begin());

	double* red_cost_d = new double[si->getNumCols()];
	double* duals_d = new double[si->getNumRows()];	// d_B' B^{-1}
	double* d_arr = new double[d.size()];
	std::copy(d.begin(), d.end(), d_arr);
	si->getReducedGradient(red_cost_d, duals_d, d_arr);	// d - d_B' B^{-1} A
	delete[] d_arr;
	delete[] duals_d;

	// Compute stability range
	for (unsigned int n = 0; n < d.size(); ++n) {
		double denum = red_cost_d[n];
		double num = red_cost_w[n];
		double stability_theta = 0.0;

//#ifdef DEBUG
		std::cout << "edge " << n << " (" << edgemap[n].first << "-"
			<< edgemap[n].second << "), coeff " << obj[n] << ", val "
			<< edgevar[n] << ", red_cost_w " << red_cost_w[n]
			<< ", red_cost_d " << red_cost_d[n] << std::endl;
//#endif

		// Three cases,
		const double num_tol = 1e-7;
		const double denum_tol = 1e-4;
		if (fabs(num) > num_tol && fabs(denum) > denum_tol) {
			//    i) fabs(denum) > 0 and num > 0
			stability_theta = - num / denum;

			// If theta >= stability_theta, then basis remains optimal.  This
			// implies that if stability_theta > 0, then all theta <=
			// stability_theta are stable, as well as if stability_theta > 0, then
			// all theta >= stability_theta are stable.
			if (stability_theta >= 0.0) {
				theta_max[n] = stability_theta;
			} else {
				theta_min[n] = stability_theta;
			}
		} else if (fabs(denum) > denum_tol) {
			//  ii) fabs(denum) > 0 and num = 0
			// the case when there are multiple optimal solutions for this
			// edge weight choice, hence any + theta d_i perturbation is going
			// to lose optimality.
			theta_min[n] = theta_max[n] = 0.0;
		} else if (fabs(num) > num_tol) {
			// iii) denum = 0, fabs(num) > 0
			// the case when d is orthogonal to the 1d subspace corresponding
			// to the n'th variable.  Any perturbation along d has no effect
			// (similar to the "iv)" case).
		} else {
			//  iv) denum = 0 and num = 0,
			// the case when d is co-linear to the constraints, that is, it is
			// stable for all theta.
		}
//#ifdef DEBUG
		std::cout << "   theta_min " << theta_min[n]
			<< ", theta_max " << theta_max[n] << std::endl;
//#endif
	}
	delete[] red_cost_d;
	si->disableSimplexInterface();
}

// Complexity is O(|E| log |V|) for one edge node1-node2.
bool GraphClustering::FindShortestPathDijkstra(
	unsigned int node1, unsigned int node2, double z,
	std::list<unsigned int>& edge_indices) const {
	return (FindShortestPathDijkstra(node1, node2, z,
		edgemap, nodemap, edgevar, edge_indices));
}

bool GraphClustering::FindShortestPathDijkstra(
	unsigned int node1, unsigned int node2, double z,
	const std::vector<std::pair<unsigned int, unsigned int> >& g_edgemap,
	const std::vector<std::set<unsigned int> >& g_nodemap,
	const std::vector<double>& g_edgedist,
	std::list<unsigned int>& edge_indices) const {
	// If this is a zero edge, we have nothing to do.  This is true in general
	// as all graph edge weights are >= 0.
	if (z < zero_one_tolerance)
		return (false);

	// Clear output path
	edge_indices.clear();

	// Initialize distance and previous array
	std::vector<double> dist(g_nodemap.size());
	std::fill(dist.begin(), dist.end(), std::numeric_limits<double>::max());
	std::vector<int> previous(g_nodemap.size());
	std::fill(previous.begin(), previous.end(), -1);	// undefined

	// Initialize priority queue for nearest front
	node_distance_heap front_queue;

	dist[node1] = 0.0;	// source has zero distance
	front_queue.push(std::pair<double, unsigned int>(0.0, node1));

	while (front_queue.empty() == false) {
		unsigned int current_node = front_queue.top().second;
		double current_dist = front_queue.top().first;
		front_queue.pop();

		// If this node has become outdated, i.e. its distance is larger than
		// the current best path to it, just remove it from the queue.
		if (current_dist > dist[current_node])
			continue;

		// Reached target node on shortest path, we can stop.
		if (current_node == node2)
			break;

		// For all neighbors of front
		const std::set<unsigned int>& adjacent_edges = g_nodemap[current_node];
		for (std::set<unsigned int>::const_iterator ei =
			adjacent_edges.begin(); ei != adjacent_edges.end(); ++ei) {
			const std::pair<unsigned int, unsigned int>& edge = g_edgemap[*ei];
			unsigned int adjacent_node = (edge.first == current_node) ?
				edge.second : edge.first;

			// New distance to reach adjacent_node over current_node's path
			assert(g_edgedist[*ei] >= 0.0);
			double alt = dist[current_node] + g_edgedist[*ei];
			if (alt < dist[adjacent_node]) {	// better than previous?
				dist[adjacent_node] = alt;
				previous[adjacent_node] = current_node;
				front_queue.push(
					std::pair<double, unsigned int>(alt, adjacent_node));
			}
		}
	}

	// There must be a path from node1 to node2 because we have an edge
//	assert(previous[node2] != -1);
	// If there is no shorter path than the direct edge with cost z, then all
	// constraints with z on the left hand side are fulfilled.
	if (dist[node2] >= (z - zero_one_tolerance))
		return (false);

	// We have a constraint violation of (dist[node2] - z) > 0, hence we can
	// generate a violated constraint from the shortest path.
	for (unsigned int cur_node = node2; cur_node != node1; ) {
		const std::set<unsigned int>& adjacent_edges = g_nodemap[cur_node];
		for (std::set<unsigned int>::const_iterator ei =
			adjacent_edges.begin(); ei != adjacent_edges.end(); ++ei) {
			const std::pair<unsigned int, unsigned int>& edge = g_edgemap[*ei];
			unsigned int adjacent_node = (edge.first == cur_node) ?
				edge.second : edge.first;

			// This edge does not belong to the shortest path
			if (static_cast<int>(adjacent_node) != previous[cur_node])
				continue;

			// It does belong to the shortest path, add it
			edge_indices.push_back(*ei);
			cur_node = previous[cur_node];
			break;
		}
	}
	return (true);
}

void GraphClustering::OutputViolationStatistics(
	const std::list<double>& violation_scores) const {
	double violation_min = *std::min_element(violation_scores.begin(),
		violation_scores.end());
	double violation_max = *std::max_element(violation_scores.begin(),
		violation_scores.end());
	double hist_binsize = (violation_max - violation_min) /
		static_cast<double>(violation_histogram_bins);
	std::vector<double> violation_histogram(violation_histogram_bins);
	std::fill(violation_histogram.begin(), violation_histogram.end(), 0.0);
	double total = 0.0;
	for (std::list<double>::const_iterator vil = violation_scores.begin();
		vil != violation_scores.end(); ++vil) {
		unsigned int bin_index =
			static_cast<unsigned int>((*vil - violation_min) / hist_binsize);
		if (bin_index >= violation_histogram.size())
			bin_index = violation_histogram.size() - 1;
		violation_histogram[bin_index] += 1.0;
		total += 1.0;
	}

	std::cout << "Constraint violations between " << violation_min
		<< " and " << violation_max << std::endl;
	for (unsigned int bin_index = 0;
		bin_index < violation_histogram.size(); ++bin_index) {
		std::cout << "   " << bin_index
			<< ": (" << (static_cast<double>(bin_index) * hist_binsize + violation_min) << " to "
			<< ((static_cast<double>(bin_index) + 1.0) * hist_binsize + violation_min) << "): "
			<< violation_histogram[bin_index] << ", "
			<< (100.0 * (violation_histogram[bin_index] / total)) << " percent"
			<< std::endl;
	}
}

bool GraphClustering::AddViolatedConstraints(OsiClpSolverInterface* si_cur,
	const std::vector<double>& edgevar_cur) {
	bool has_ineq = false;
	has_ineq = SeparateCycleInequalities(si_cur, edgevar_cur);

	bool is_fractional = IsFractionalSolution(edgevar_cur);
	if (has_ineq == false && is_fractional) {
		if (do_separate_oddwheel)
			has_ineq = SeparateOddWheelInequalities(si_cur, edgevar_cur);
	} else if (has_ineq == false) {
		std::cout << "Solution satisfies cycle inequalities and is integral, "
			<< "hence optimal." << std::endl;
	}

	return (has_ineq);
}

bool GraphClustering::SeparateCycleInequalities(OsiClpSolverInterface* si_cur,
	const std::vector<double>& edgevar_cur) {
	std::cout << std::endl;
	std::cout << "Solving cycle inequalities separation problem" << std::endl;
	boost::progress_display show_progress(edgevar_cur.size());

	// For histogram statistics, keep all violation scores
	std::list<double> violation_scores;

	// For each edge, try to find a shorter path around it
	unsigned int violated_count = 0;
	for (unsigned int ei = 0; ei < edgevar_cur.size(); ++ei) {
		++show_progress;

		std::list<unsigned int> edge_indices;
		bool violation = FindShortestPathDijkstra(edgemap[ei].first,
			edgemap[ei].second, edgevar_cur[ei], edgemap, nodemap, edgevar_cur,
			edge_indices);
		// FIXME: edgevar
		if (violation == false)
			continue;

		// Found a violated constraint, add it (compute violation on the way)
		double violation_magnitude = edgevar_cur[ei];
		CoinPackedVector row;
		row.insert(ei, 1.0);
#if 0
		std::cout << "  edge: z_{"
			<< edgemap[ei].first << "," << edgemap[ei].second
			<< "} <= ";
#endif
		for (std::list<unsigned int>::const_iterator pi = edge_indices.begin();
			pi != edge_indices.end(); ++pi) {
			row.insert(*pi, -1.0);
#if 0
			std::cout << " + z_{" << edgemap[*pi].first << ","
				<< edgemap[*pi].second << "} ";
#endif
			violation_magnitude -= edgevar_cur[*pi];
		}
#if 0
		std::cout << std::endl;
#endif
		si_cur->addRow(row, -si_cur->getInfinity(), 0.0);
		violation_scores.push_back(violation_magnitude);
#if 0
		std::cout << "    added violated constraint, edge " << ei
			<< " (" << edgemap[ei].first << "-" << edgemap[ei].second << "), "
			<< "violation " << violation_magnitude << std::endl;
#endif
		violated_count += 1;
		if (violated_count >= maximum_constraints_per_iter)
			break;
	}
	std::cout << "   added " << violated_count << " violated constraints."
		<< std::endl;

	// Print some statistics and a histogram
	if (violated_count > 0)
		OutputViolationStatistics(violation_scores);

	return (violated_count > 0);
}

bool GraphClustering::SeparateOddWheelInequalities(OsiClpSolverInterface* si_cur,
	const std::vector<double>& edgevar_cur) {
	std::cout << std::endl;
	std::cout << "Solving odd wheel inequality separation problem" << std::endl;
	boost::progress_display show_progress(nodemap.size());
	std::vector<double> violation_scores;

	// For each vertex vi = i_0 and odd cycle C \subseteq E \setminus
	// \delta(i_0), try to identify odd wheel inequalities of the form
	//   \sum_{m \in C} z_m - \sum_{i \in V(C)} z_{i0,i} \leq \frac{|C|-1}{2}.
	//   "outer wheel"      - "inner spokes"
	//
	// The intuition behind odd wheel inequalities is roughly the following:
	// "If vertices on the cycle are mostly in different clusters, the inner
	//  vertex can't be in too many of their clusters at once."
	unsigned int violated_count = 0;
	for (unsigned int vi = 0; vi < nodemap.size(); ++vi) {
		++show_progress;

		// 1. Collect set vertices adjacent to vi and their spokes
		std::vector<unsigned int> adj;	// small_to_big[adj_index] = G_index
		__gnu_cxx::hash_set<unsigned int> hadj;
		std::vector<unsigned int> adj_spokes;	// edge id's of spokes
		// vertex map: big_to_small[G_index] = adj_index.
		std::map<unsigned int, unsigned int> big_to_small;
		for (std::set<unsigned int>::const_iterator
			ei = nodemap[vi].begin(); ei != nodemap[vi].end(); ++ei) {
			// add the vertex id at the other end of the edge
			unsigned int ei_adj = (edgemap[*ei].first == vi) ?
				edgemap[*ei].second : edgemap[*ei].first;
			adj.push_back(ei_adj);	// cycle vertex id
			hadj.insert(ei_adj);
			big_to_small[ei_adj] = adj_spokes.size();
			adj_spokes.push_back(*ei);	// spoke edge id in the big graph
		}

		// 2. Setup "minimum weight odd circuit" problem graph from adj.
		std::vector<double> oc_edgedist;
		std::vector<unsigned int> oc_edge_small_to_big;
		std::vector<std::pair<unsigned int, unsigned int> > oc_edgemap;
		std::vector<std::set<unsigned int> > oc_nodemap(2 * adj.size());

		// For each vertex i in adj, we introduce two vertices i' and i'' in
		// the new graph.  These are not connected.
		// For each edge (i,j) \in E, with i \in adj, j \in adj, we introduce
		// two edges (i',j''), (i'',j') into the graph.  The edge distance
		// w_{i',j''} = w_{i'',j'} is computed as
		//   w_{i',j''} = 0.5 - z_{i,j} + 0.5*(z_{vi,i} + z_{vi,j}).

		//     i) collect all edges from E that are within "adj \times adj"
		std::list<unsigned int> adj_times_adj_edges;
		for (unsigned int ei = 0; ei < edgemap.size(); ++ei) {
			// Check whether the edge is in adj \times adj
			if (hadj.find(edgemap[ei].first) == hadj.end() ||
				hadj.find(edgemap[ei].second) == hadj.end())
				continue;

			// Indices of the adjacent nodes in the new graph
			unsigned int small_first = big_to_small[edgemap[ei].first];
			unsigned int small_second = big_to_small[edgemap[ei].second];

			// The edge is present, compute the Deza score [Deza1992]
			// The score should always be non-negative if all cycle
			// inequalities have been added.
			double w_ij = 0.5 - edgevar_cur[ei] + 0.5 * (
				edgevar_cur[adj_spokes[small_first]] +
				edgevar_cur[adj_spokes[small_second]]);
			assert(w_ij >= -1e-8);
			if (w_ij < 0.0)
				w_ij = 0.0;

			// The following addition is very subtle: if not done, a cycle
			// such as the following one could be found.
			//
			//        .-->>--.     .-->>--.     .-->>--.
			//    (0)          (1)          (2)          (3)
			//        '-->>--'     '--<<-----------<<--'
			//
			// While the corresponding 5-wheel produces a valid inequality,
			// we should instead add the 3-wheel consisting of (1), (2) and
			// (3).  This is achieved implicitly by slightly penalizing each
			// edge by 1e-10, thus producing a preference towards the shorter,
			// but equal cost cycle (1), (2), (3).
			//
			// XXX: we now do explicit inner cycle detection
			// w_ij += 1e-10;

			// Add the edge twice with equal costs
			oc_nodemap[2*small_first+0].insert(oc_edgemap.size());
			oc_nodemap[2*small_second+1].insert(oc_edgemap.size());
			oc_edgemap.push_back(std::pair<unsigned int, unsigned int>(
				2*small_first+0, 2*small_second+1));
			oc_edgedist.push_back(w_ij);
			oc_edge_small_to_big.push_back(ei);

			oc_nodemap[2*small_first+1].insert(oc_edgemap.size());
			oc_nodemap[2*small_second+0].insert(oc_edgemap.size());
			oc_edgemap.push_back(std::pair<unsigned int, unsigned int>(
				2*small_first+1, 2*small_second+0));
			oc_edgedist.push_back(w_ij);
			oc_edge_small_to_big.push_back(ei);
		}

		//   iii) solve the odd circuit problem for each start/end vertex,
		//        keeping the minimal one
		double max_violation = -1.0;
		oddwheel_inequality_heap max_ineq;
		for (unsigned int n = 0; n < adj.size(); ++n) {
			std::list<unsigned int> edge_indices;
			bool has_violation_wheel = FindShortestPathDijkstra(
				2*n + 0, 2*n + 1, 0.5, oc_edgemap, oc_nodemap, oc_edgedist,
				edge_indices);

			// No violation, keep searching
			if (has_violation_wheel == false)
				continue;

			// Measure violation, keep highest
			double violation = -0.5;
			__gnu_cxx::hash_set<unsigned int> inner_cycle_collider;
			bool c_vertex_collides = false;
			for (std::list<unsigned int>::const_iterator
				eii = edge_indices.begin(); eii != edge_indices.end(); ++eii) {

				// Check all inner cycle vertices for internal cycles
				if (eii != edge_indices.begin()) {
					unsigned int c_vertex = edgemap[oc_edge_small_to_big[*eii]].first;
					c_vertex_collides = inner_cycle_collider.find(c_vertex)
						!= inner_cycle_collider.end();
					if (c_vertex_collides)
						break;
					inner_cycle_collider.insert(c_vertex);
				}
				violation += oc_edgedist[*eii];
			}
			if (c_vertex_collides)
				continue;

			violation *= -1.0;
			assert(violation >= 0.0);

			max_ineq.push(ViolatedOddWheelInequality(violation,
				edge_indices, oc_edge_small_to_big));
		}

		// No violated inequality found, continue
		if (max_ineq.empty())
			continue;

		unsigned int added = 0;
		while (max_ineq.empty() == false) {
			if (added > 50)	// FIXME
				break;

			ViolatedOddWheelInequality ineq = max_ineq.top();
			max_ineq.pop();

			const std::list<unsigned int>& max_edge_indices = ineq.edge_indices;
			const std::vector<unsigned int>& max_oc_edge_small_to_big =
				ineq.oc_edge_small_to_big;

			// Setup odd wheel inequality related to vi
			CoinPackedVector row;
			std::set<unsigned int> cycle_vertices;
			for (std::list<unsigned int>::const_iterator
				pi = max_edge_indices.begin(); pi != max_edge_indices.end(); ++pi) {
				// *pi is an edge index in the small search graph we created.  We
				// first map it back to original indices
				assert(*pi >= 0 && *pi < max_oc_edge_small_to_big.size());
				unsigned int G_ei = max_oc_edge_small_to_big[*pi];
				assert(G_ei >= 0 && G_ei < edgemap.size());
				row.insert(G_ei, 1.0);	// cycle edge

				// Vertices on the cycle
				cycle_vertices.insert(edgemap[G_ei].first);
				cycle_vertices.insert(edgemap[G_ei].second);
			}
			// Add spokes
			for (std::set<unsigned int>::const_iterator cvi = cycle_vertices.begin();
				cvi != cycle_vertices.end(); ++cvi) {
#ifdef DEBUG
				std::cout << "        cycle edge " <<
					adj_spokes[big_to_small[*cvi]] << std::endl;
#endif
				row.insert(adj_spokes[big_to_small[*cvi]], -1.0);
			}

			double rhs = 0.5 * (static_cast<double>(max_edge_indices.size()) - 1.0);
			if (doing_stability_analysis) {
				row.insert(edgevar_cur.size(), rhs);	// Az + b alpha <= 0
				rhs = 0.0;
			}
			// Add to LP the violated inequality
			si_cur->addRow(row, -si_cur->getInfinity(), rhs);
			added += 1;

			violation_scores.push_back(max_violation);
#ifdef DEBUG
			std::cout << "    added odd wheel violated constraint, vertex " << vi << ", "
				<< "violation " << ineq.violation << std::endl;
#endif
			violated_count += 1;
			if (violated_count >= maximum_constraints_per_iter)
				break;
		}
	}

	if (violated_count > 0) {
		std::cout << "Added " << violated_count
			<< " odd wheel inequalities." << std::endl;
	}

	return (violated_count > 0);
}

// Kernighan & Lin heuristic, adapted for the unconstrained GPP
double GraphClustering::SolveHeuristicallyKL() {
	// Current partitioning
	std::vector<__gnu_cxx::hash_set<unsigned int> > parts;

	// Start with just a single partition
	__gnu_cxx::hash_set<unsigned int> part_init(nodemap.size());
	for (unsigned int ni = 0; ni < nodemap.size(); ++ni)
		part_init.insert(ni);
	parts.push_back(part_init);

	bool seen_change = true;
	double total_obj = 0.0;
	for (unsigned int iter = 1; seen_change; ++iter) {
		std::cout << "KL iter " << iter
			<< ", partition count " << parts.size()
			<< ", obj " << total_obj << std::endl;
		seen_change = false;
		std::set<unsigned int> prune_partition_indices;
		std::vector<__gnu_cxx::hash_set<unsigned int> > new_partitions;

		// Try to improve all pairwise partitions
		for (unsigned int part1 = 0; part1 < parts.size(); ++part1) {
			for (unsigned int part2 = part1 + 1; part2 < parts.size(); ++part2) {
				double cur_delta = SolveHeuristicallyKLTwoPartitions(parts[part1],
					parts[part2]);
				if (cur_delta < -1e-10)
					seen_change = true;
				total_obj += cur_delta;
			}

			// Partition went bust: prune
			if (parts[part1].size() == 0) {
				prune_partition_indices.insert(part1);
				std::cout << "   -> pruned one partition" << std::endl;
				continue;
			}

			// If there is more than one element left, try to split
			if (parts[part1].size() <= 1)
				continue;

			__gnu_cxx::hash_set<unsigned int> part_empty(parts[part1].size());
			double empty_delta = SolveHeuristicallyKLTwoPartitions(parts[part1],
				part_empty);
			if (part_empty.size() == 0)
				continue;	// Nothing split off

			seen_change = true;
			std::cout << "   -> created new partition of size " << part_empty.size() << std::endl;
			new_partitions.push_back(part_empty);
			total_obj += empty_delta;
		}

		// Prune partitions, then append new ones
		for (std::set<unsigned int>::const_reverse_iterator pri =
			prune_partition_indices.rbegin();
			pri != prune_partition_indices.rend(); ++pri)
			parts.erase(parts.begin() + *pri);
		parts.insert(parts.end(), new_partitions.begin(), new_partitions.end());
	}

	// Set internal state correctly
	__gnu_cxx::hash_map<unsigned int, unsigned int> node_to_partition_id;
	for (unsigned int partition_id = 0; partition_id < parts.size(); ++partition_id) {
		for (__gnu_cxx::hash_set<unsigned int>::const_iterator
			pi = parts[partition_id].begin(); pi != parts[partition_id].end(); ++pi)
			node_to_partition_id[*pi] = partition_id;
	}
	for (unsigned int ei = 0; ei < edgemap.size(); ++ei) {
		assert(node_to_partition_id.find(edgemap[ei].first) != node_to_partition_id.end());
		assert(node_to_partition_id.find(edgemap[ei].second) != node_to_partition_id.end());
		edgevar[ei] = node_to_partition_id[edgemap[ei].first] ==
			node_to_partition_id[edgemap[ei].second] ? 0.0 : 1.0;
	}

	// Setup theta_path array (we cannot really compute anything related to
	// stability).
	theta_path.clear();
	theta_path.push_back(std::pair<double, double>(0.0, 0.0));
	theta_iters_exceeded.clear();
	theta_iters_exceeded.push_back(false);

	return (total_obj);
}

// Kernighan & Lin heuristic subproblem: optimize interchanges between two
// partitions only.  One of the two partitions is allowed to be empty (this
// way new partitions can be created).
//
// Return the total objective delta (always <= 0.0, hopefully negative).
double GraphClustering::SolveHeuristicallyKLTwoPartitions(
	__gnu_cxx::hash_set<unsigned int>& part1,
	__gnu_cxx::hash_set<unsigned int>& part2) const {
	double total_gain = 0.0;
	for (unsigned int iter = 1; true; ++iter) {
		std::cout << "ITER " << iter << ", total_gain " << total_gain << std::endl;
#if 0
		std::cout << "   PART1:";
		for (__gnu_cxx::hash_set<unsigned int>::const_iterator pi1 = part1.begin();
			pi1 != part1.end(); ++pi1)
			std::cout << " " << *pi1;
		std::cout << std::endl;
		std::cout << "   PART2:";
		for (__gnu_cxx::hash_set<unsigned int>::const_iterator pi2 = part2.begin();
			pi2 != part2.end(); ++pi2)
			std::cout << " " << *pi2;
		std::cout << std::endl;
#endif

		// D[n] = E_n - I_n
		__gnu_cxx::hash_map<unsigned int, double> D(part1.size() + part2.size());

		// 1. Compute initial D values
		__gnu_cxx::hash_set<unsigned int> part_union;
		part_union.insert(part1.begin(), part1.end());
		part_union.insert(part2.begin(), part2.end());

		for (__gnu_cxx::hash_set<unsigned int>::const_iterator
			pi = part_union.begin(); pi != part_union.end(); ++pi) {
			__gnu_cxx::hash_set<unsigned int>& part =
				part1.find(*pi) != part1.end() ? part1 : part2;
			__gnu_cxx::hash_set<unsigned int>& part_other =
				part1.find(*pi) != part1.end() ? part2 : part1;

			// Compute E_a, I_a
			double E_a = 0.0;
			double I_a = 0.0;
			// For each edge
			for (std::set<unsigned int>::const_iterator ei = nodemap[*pi].begin();
				ei != nodemap[*pi].end(); ++ei) {
				unsigned int node_other = (*pi == edgemap[*ei].first) ?
					edgemap[*ei].second : edgemap[*ei].first;

				if (part.find(node_other) != part.end()) {
					// Change if this internal edge is made external
					I_a += obj[*ei];
				} else if (part_other.find(node_other) != part_other.end()) {
					// Change if this external edge is made internal
					E_a += obj[*ei];
				}
			}
			// Compute difference: the gain for moving a to the other set
			D[*pi] = E_a - I_a;
		}

		// Set of already moved nodes.  These are not considered when the maximum
		// exchange is sought.
		__gnu_cxx::hash_set<int> moved_nodes(std::max(part1.size(), part2.size()));

		// Individual gains
		std::vector<double> gain_sequence;

		// Running sum and cummulative gain sums
		double gain_sum = 0.0;
		std::vector<double> gain_sum_sequence;
		gain_sum_sequence.push_back(gain_sum);	// zero moves: zero gain
		double max_gain_sum = 1e-10;	// minimum required gain
		unsigned int max_gain_moves = 0;

		// Sequence of greedy best moves
		std::vector<std::pair<int, int> > move_sequence;

		// 2. Build greedy move sequence
		do {
			// 2. Choose maximum gain greedily
			double max_gain = -std::numeric_limits<double>::infinity();
			int part1_node = 0;
			int part2_node = 0;
			bool swap_possible = false;
			for (__gnu_cxx::hash_set<unsigned int>::const_iterator pi1 = part1.begin();
				pi1 != part1.end(); ++pi1) {
				// Has the node already been considered?  If so, skip
				if (moved_nodes.find(*pi1) != moved_nodes.end())
					continue;

				// First consider the simple "move to part2" move, without
				// moving anything back
				swap_possible = true;
				if (D[*pi1] > max_gain) {
					max_gain = D[*pi1];
					part1_node = *pi1;
					part2_node = -1;	// 'empty'
				}

				for (__gnu_cxx::hash_set<unsigned int>::const_iterator pi2 = part2.begin();
					pi2 != part2.end(); ++pi2) {
					// Has the node already been considered?  If so, skip
					if (moved_nodes.find(*pi2) != moved_nodes.end())
						continue;

					// Compute gain achieved: D_1 + D_2 - 2w(pi1,pi2)
					double cur_gain = D[*pi1] + D[*pi2];
					for (std::set<unsigned int>::const_iterator ei = nodemap[*pi1].begin();
						ei != nodemap[*pi1].end(); ++ei) {
						if (edgemap[*ei].first != *pi2 && edgemap[*ei].second != *pi2)
							continue;

						cur_gain -= 2.0 * obj[*ei];
						break;
					}
					// Track maximum
					if (cur_gain <= max_gain)
						continue;
					max_gain = cur_gain;
					part1_node = *pi1;
					part2_node = *pi2;
				}
			}

			// Also consider simple "move to part1" moves, without moving
			// anything from part1
			for (__gnu_cxx::hash_set<unsigned int>::const_iterator pi2 = part2.begin();
				pi2 != part2.end(); ++pi2) {
				// Has the node already been considered?  If so, skip
				if (moved_nodes.find(*pi2) != moved_nodes.end())
					continue;

				swap_possible = true;
				if (D[*pi2] > max_gain) {
					max_gain = D[*pi2];
					part1_node = -1;	// 'empty'
					part2_node = *pi2;
				}
			}

			// No more movements possible, abort
			if (swap_possible == false)
				break;

			moved_nodes.insert(static_cast<int>(part1_node));
			moved_nodes.insert(static_cast<int>(part2_node));
			gain_sequence.push_back(max_gain);
			gain_sum += max_gain;
			gain_sum_sequence.push_back(gain_sum);
			if (gain_sum > max_gain_sum) {
				max_gain_sum = gain_sum;
				// Store number of moves to make from move_sequence to achieve
				// current max_gain
				max_gain_moves = gain_sum_sequence.size() - 1;
			}
			move_sequence.push_back(std::pair<int, int>(part1_node, part2_node));

			// Recompute D for part 1
			if (part1_node != -1) {
				for (std::set<unsigned int>::const_iterator ei = nodemap[part1_node].begin();
					ei != nodemap[part1_node].end(); ++ei) {
					unsigned int node_other =
						(edgemap[*ei].first == static_cast<unsigned int>(part1_node)) ?
						edgemap[*ei].second : edgemap[*ei].first;

					if (part1.find(node_other) != part1.end()) {
						// Other node is also in part1
						D[node_other] += 2.0 * obj[*ei];
					} else if (part2.find(node_other) != part2.end()) {
						// Other node is in part2
						D[node_other] -= 2.0 * obj[*ei];
					}
				}
			}
			// Recompute D for part 2
			if (part2_node != -1) {
				for (std::set<unsigned int>::const_iterator ei = nodemap[part2_node].begin();
					ei != nodemap[part2_node].end(); ++ei) {
					unsigned int node_other =
						(edgemap[*ei].first == static_cast<unsigned int>(part2_node)) ?
						edgemap[*ei].second : edgemap[*ei].first;

					if (part2.find(node_other) != part2.end()) {
						// Other node is also in part2
						D[node_other] += 2.0 * obj[*ei];
					} else if (part1.find(node_other) != part1.end()) {
						// Other node is in part1
						D[node_other] -= 2.0 * obj[*ei];
					}
				}
			}
		} while (true);

		// 3. Perform optimal greedy move sequence, if it yields improvement
		if (max_gain_moves == 0)
			break;

		for (unsigned int mi = 0; mi < max_gain_moves; ++mi) {
			int node1 = move_sequence[mi].first;
			int node2 = move_sequence[mi].second;
			if (node1 != -1) {
				part1.erase(node1);
				part2.insert(node1);
			}
			if (node2 != -1) {
				part2.erase(node2);
				part1.insert(node2);
			}
		}
		total_gain += max_gain_sum;
	}

	return (-total_gain);	// FIXME: negative is correct?
}

}


