/* Linear Program Graph Clustering
 *
 * Copyright (C) 2008 -- Sebastian Nowozin <nowozin@gmail.com>
 */

#ifndef GRAPHCLUSTERING_H
#define GRAPHCLUSTERING_H

#include <vector>
#include <ext/hash_set>
#include <set>
#include <list>
#include <queue>
#include <utility>
#include <boost/lambda/lambda.hpp>

#include <gmm/gmm.h>

#include <coin/OsiClpSolverInterface.hpp>
#include <coin/CoinPackedMatrix.hpp>
#include <coin/CoinPackedVector.hpp>

namespace Clustering {

using namespace boost::lambda;

class GraphClustering {
public:
	// do_separate_oddwheel: Whether to use odd-wheel inequalities.
	// max_iter: if <= 0, allow an unlimited number of relaxed LP iterations,
	//    if > 0, limit it to the given number.
	GraphClustering(bool do_separate_oddwheel = true,
		int max_iter = -1);
	~GraphClustering();

	// Initialize using a given symmetric weight matrix.  The weight matrix
	// may be normalized.  The lower triangular part is ignored, so only the
	// diagonal and triu part is used.
	// If the matrix is truly sparse only the given matrix elements are
	// instantiated as edges.  If the matrix contains zero elements these are
	// also instantiated as edges.  That is, a weight which does not exist in
	// K produces no edge, but if the weight exist it creates an edge, no
	// matter what its value is.
	// If minimum or maximum cluster sizes are given >= 1, additional
	// constraints will be added a-priori.  This is allows only for the
	// completely connected case.
	//
	// The second method is the same but for dense matrices.  Here, even
	// elements with zero weights are explicitly instantiated.
	void Initialize(const gmm::row_matrix<gmm::wsvector<double> >& K);
	void Initialize(const gmm::dense_matrix<double>& K);

	// Solve for the solution path using a given linear shift range applied to
	// all elements in the matrix K (zero elements are ignored).
	// For theta_left, the value -std::numeric_limits<double>::infinity() is
	// valid and denotes the extreme of each sample being one cluster, for
	// theta_right, the value std::numeric_limits<double>::infinity() is valid
	// and denotes the other extreme of one joint cluster with all samples.
	// The path is tracked from right to left.
	//
	// Usually you would want to call the method as
	//     Solve(-2.4, std::numeric_limits<double>::infinity()),
	// where -2.4 is just some constant.  This will produce clusterings
	// starting with one cluster.
	void Solve(double theta_left = 0.0, double theta_right = 0.0);

	double Objective() const;

	// The signed objective function coefficients
	const std::vector<double>& ObjectiveCoefficients() const;

	// Return edge variables in the same order as the Edgemap() order.
	// This is the function to use for a theta range of [0,0], where only one
	// solution is computed.
	const std::vector<double>& Solution() const;

	// Return entire solution path, theta ranges and solutions.
	const std::vector<std::pair<double, double> >& ThetaPath() const;
	const std::vector<std::vector<double> >& ThetaSolutions() const;
	// Return flags for each solution whether the number of maximum iterations
	// was exceeded.  If this is the case, then the solution may be suboptimal
	// even though it is integral.
	const std::vector<bool>& ThetaItersExceeded() const;

	// Return true if the current solution has fractional elements, false
	// otherwise.  If false is returned, the solution is also a globally
	// optimal integral solution.
	bool IsFractionalSolution() const;

	// Return true if solution is fractional, false if integral.
	bool IsFractionalSolution(const std::vector<double>& ev) const;

	const std::vector<std::pair<unsigned int, unsigned int> >& Edgemap() const;

	// Global range sensitivity analysis
	// For a given direction d, compute the cost vector stability range, i.e.
	// the range of \theta \in \R, such that the current optimal solution is
	// still optimal for (c + \theta d).
	//
	// Note that theta_min may be -infinity and theta_max may be +infinity.
	// Also, they may be (0, 0), in which there are multiple optimal
	// solutions, eg. the the current cost vector is normal to a strictly
	// active constraint.
	//
	// Also note that the returned range is guaranteed, however the reverse is
	// not true: even outside the stability range the current solution may be
	// optimal.
	void ComputeStabilityRange(const std::vector<double>& d,
		double& theta_min, double& theta_max);

	// Alternative, more robust sensitivity analysis using sub-LP
	void ComputeStabilityRangeLP(const std::vector<double>& d,
		double& theta_min, double& theta_max);

	// General sensitivity analysis.
	//
	// Compute a stability box around each edge weight.  Given a currently
	// optimal solution, and a cost direction d, compute for each edge cost an
	// individual stability range for which the current solution/basis remains
	// optimal.
	//   Note that the for each variable e_j, the stability range is of the
	// form [-infinity; a] with a >= 0, or [a; infinity] with a <= 0, or
	// [0; 0].
	void ComputeStabilityBox(const std::vector<double>& d,
		std::vector<double>& theta_min, std::vector<double>& theta_max);

	// Solve heuristically using Kernighan-Lin heuristic.
	//
	// The heuristic is adapted from
	//     B.W. Kernighan, S. Lin,
	//     "An Efficient Heuristic Procedure for Partitioning Graphs",
	//     1970, AT&T.
	//
	// The number of partitions is also optimized over and the final solution
	// returned is always feasible and always a local optima with respect to
	// the heuristics' search neighborhood definition (which is quite large).
	// The implementation is not particularly optimized but should be fast
	// enough to partition graphs of size a few million edges.
	//
	// The heuristic could be parallelized in this method: all pairwise
	// optimizations can be grouped and carried out in parallel across groups,
	// serially within.
	//
	// Return objective.
	double SolveHeuristicallyKL();

	// Solve heuristically for only two partitions.  This method is used as a
	// subroutine for the KL heuristic and the real work happens here.
	double SolveHeuristicallyKLTwoPartitions(
		__gnu_cxx::hash_set<unsigned int>& part1,
		__gnu_cxx::hash_set<unsigned int>& part2) const;

protected:
	// The linear program model
	OsiClpSolverInterface* si;
	unsigned int dynamic_constraint_start;

	// Whether to separate odd-wheel inequalities
	bool do_separate_oddwheel;

	// Maximum number of outer relaxed-LP iterations
	int max_iter;

	// edgemap: edge_index -> (node_index_1, node_index_2)
	std::vector<std::pair<unsigned int, unsigned int> > edgemap;
	// nodemap: node_index -> set of edge_indices
	std::vector<std::set<unsigned int> > nodemap;
	// edgevar: edge_index -> [0, 1]
	std::vector<double> edgevar;

	// The vanilla objective as given through K
	std::vector<double> obj;

	// Stability ranges and respective optimal clustering solutions
	std::vector<std::pair<double, double> > theta_path;
	std::vector<std::vector<double> > theta_edgevars;
	std::vector<bool> theta_iters_exceeded;

	static const unsigned int maximum_constraints_per_iter = 50000;
	static const double zero_one_tolerance = 1e-8;
	static const unsigned int violation_histogram_bins = 10;

	struct compare_distance_node_pair {
		bool operator()(const std::pair<double, unsigned int>& v1,
			const std::pair<double, unsigned int>& v2) const {
			return (v1.first > v2.first);
		}
	};
	typedef std::priority_queue<
		std::pair<double, unsigned int>,
		std::vector<std::pair<double, unsigned int> >,
		compare_distance_node_pair
	> node_distance_heap;

	// Heap for violated odd wheel inequalities
	struct ViolatedOddWheelInequality {
		ViolatedOddWheelInequality(double violation,
			const std::list<unsigned int>& edge_indices,
			const std::vector<unsigned int>& oc_edge_small_to_big)
			: violation(violation), edge_indices(edge_indices),
				oc_edge_small_to_big(oc_edge_small_to_big) {
		}

		double violation;
		std::list<unsigned int> edge_indices;
		std::vector<unsigned int> oc_edge_small_to_big;
	};
	struct compare_violated_odd_wheel_inequality {
		bool operator()(const ViolatedOddWheelInequality& v1,
			const ViolatedOddWheelInequality& v2) const {
			return (v1.violation < v2.violation);
		}
	};
	typedef std::priority_queue<
		ViolatedOddWheelInequality,
		std::vector<ViolatedOddWheelInequality>,
		compare_violated_odd_wheel_inequality
	> oddwheel_inequality_heap;

	// Critical variable for odd-wheel inequalities in auxiliary LP.
	bool doing_stability_analysis;

	// Internal initialization method: setup LP solver
	void InternalInitialize();

	// Solve the current instance, i.e. perform constraint generation and
	// iterate solving the LP
	// Return the number of outer iterations taken (constraint generation
	// iterations).
	unsigned int SolveFixed();

	// TODO: doc
	unsigned int RemoveInactiveConstraints();

	// Find shortest path between node1 and node2, using the current values in
	// edgevar as distances.  If a path strictly shorter than z is found, the
	// edge indices are returned, in path order from node2 to node1.
	//
	// Return true if a violating path has been found, false otherwise.  The
	// path is returned from node1 to node2 as a list of edge indices in
	// edge_indices.
	bool FindShortestPathDijkstra(
		unsigned int node1, unsigned int node2, double z,
		std::list<unsigned int>& edge_indices) const;

	// Same as FindShortestPathDijkstra, but the graph is completely specified
	// through the additional parameters g_edgemap, g_nodemap and g_edgevar.
	// The edge weights/distances are given in g_edgedist and must be
	// non-negative.
	bool FindShortestPathDijkstra(
		unsigned int node1, unsigned int node2, double z,
		const std::vector<std::pair<unsigned int, unsigned int> >& g_edgemap,
		const std::vector<std::set<unsigned int> >& g_nodemap,
		const std::vector<double>& g_edgedist,
		std::list<unsigned int>& edge_indices) const;

	// Print a histogram of violation values
	void OutputViolationStatistics(
		const std::list<double>& violation_scores) const;

	// Find and add violated valid inequalities for the multicut polytope.
	// Return true if violated constraints have been found, false otherwise.
	bool AddViolatedConstraints(OsiClpSolverInterface* si_cur,
		const std::vector<double>& edgevar_cur);

	// Solve the separation problem for cycle inequalities and add violated
	// ones to the LP.
	bool SeparateCycleInequalities(OsiClpSolverInterface* si_cur,
		const std::vector<double>& edgevar_cur);

	// Solve the separation problem for odd wheel inequalities and add
	// violated ones to the LP.
	bool SeparateOddWheelInequalities(OsiClpSolverInterface* si_cur,
		const std::vector<double>& edgevar_cur);

	// Return true if dual is unbounded (infinite range, primal infeasible),
	// return false if solved to optimality.
	// Return: 0 if range was computed, -1 if dual infeasible (primal
	// unbounded), the range value is \infty, 1 if primal infeasible (d
	// 	infeasible), return false if solved to optimality.
	int ComputeStabilityRangeLPSingle(OsiClpSolverInterface* si2,
		std::vector<double>& edgevar2, unsigned int alpha_index);
};

}

#endif

