
#include <cppunit/BriefTestProgressListener.h>
#include <cppunit/CompilerOutputter.h>
#include <cppunit/extensions/TestFactoryRegistry.h>
#include <cppunit/TestResult.h>
#include <cppunit/TestResultCollector.h>
#include <cppunit/TestRunner.h>
#include <cppunit/config/SourcePrefix.h>

#include <boost/random.hpp>

#include <vector>
#include <iostream>
#include <algorithm>
#include <numeric>
#include <limits>
#include <stdlib.h>

#include <coin/OsiClpSolverInterface.hpp>
#include <coin/CoinPackedMatrix.hpp>
#include <coin/CoinPackedVector.hpp>

#include "GraphClustering.h"
#include "GraphClustering_test.h"

CPPUNIT_TEST_SUITE_REGISTRATION(GraphClusteringTest);

// Simple test
void GraphClusteringTest::Simple() {
	gmm::row_matrix<gmm::wsvector<double> > K(6, 6);

	// Normalized similarity matrix
	//
	// 0.388889 0.388889 0.388889 -0.611111 -0.611111 1.388889 
	// 0.388889 0.388889 0.388889 -0.611111 -0.611111 -0.611111 
	// 0.388889 0.388889 0.388889 -0.611111 -0.611111 -0.611111 
	// -0.611111 -0.611111 -0.611111 0.388889 0.388889 0.388889 
	// -0.611111 -0.611111 -0.611111 0.388889 0.388889 0.388889 
	// 1.388889 -0.611111 -0.611111 0.388889 0.388889 0.388889 
	K(0, 0) = K(0, 1) = K(0, 2) = 0.388889;
	K(0, 3) = K(0, 4) = -0.611111;
	K(0, 5) = 1.388889;
	K(1, 1) = K(1, 2) = 0.388889;
	K(1, 3) = K(1, 4) = K(1, 5) = -0.611111;
	K(2, 2) = 0.388889;
	K(2, 3) = K(2, 4) = K(2, 5) = -0.611111;
	K(3, 3) = K(3, 4) = K(3, 5) = 0.388889;
	K(4, 4) = K(4, 5) = 0.388889;
	K(5, 5) = 0.388889;

	Clustering::GraphClustering gc;
	gc.Initialize(K);
	gc.Solve();

	// Check solution
	const std::vector<double>& sol = gc.Solution();
	const std::vector<std::pair<unsigned int, unsigned int> >& emap =
		gc.Edgemap();
	for (unsigned int ei = 0; ei < sol.size(); ++ei) {
		CPPUNIT_ASSERT_DOUBLES_EQUAL(sol[ei], (
			(emap[ei].first <= 2 && emap[ei].second >= 3 ||
				emap[ei].first >= 3 && emap[ei].second <= 2) ? 1 : 0), 1e-5);
	}

	CPPUNIT_ASSERT(true);
}

void GraphClusteringTest::SimpleSparse() {
	gmm::row_matrix<gmm::wsvector<double> > K(6, 6);

	// Normalized similarity matrix
	//
	//    -  0.1  0.1 -0.2     -     -
	//  0.1    - -0.05    -     -     -
	//  0.1 -0.1    -    -     -     -
	// -0.2    -    -    -   0.1   0.1
	//    -    -    -  0.1     - -0.05
	//    -    -    -  0.1 -0.05     -
	//
	// corresponding to the weighted graph:
	//
	//   (2)                          (4)
	//   |  \                        /  |
	//   |  0.1                    0.1  |
	//   |     \                  /     |
	// -0.05   (0)---- -0.2 ----(3)   -0.05
	//   |     /                  \     |
	//   |  0.1                    0.1  |
	//   |  /                        \  |
	//   (1)                          (5)
	//
	// which should have the optimal partitioning: {0,1,2} {3,4,5}
	K(0, 1) = 0.1;
	K(0, 2) = 0.2;
	K(0, 3) = -0.2;
	K(1, 2) = -0.05;
	K(3, 4) = 0.1;
	K(3, 5) = 0.1;
	K(4, 5) = -0.05;

	Clustering::GraphClustering gc;
	gc.Initialize(K);
	gc.Solve();

	// Check solution
	const std::vector<double>& sol = gc.Solution();
	std::cout << "solution: " << gmm::row_vector(sol) << std::endl;
	const std::vector<std::pair<unsigned int, unsigned int> >& emap =
		gc.Edgemap();
	for (unsigned int ei = 0; ei < sol.size(); ++ei) {
#ifdef DEBUG
		std::cout << "edge " << emap[ei].first << "-" << emap[ei].second
			<< ": " << sol[ei] << std::endl;
#endif
		CPPUNIT_ASSERT_DOUBLES_EQUAL(sol[ei], (
			(emap[ei].first <= 2 && emap[ei].second >= 3 ||
				emap[ei].first >= 3 && emap[ei].second <= 2) ? 1 : 0), 1e-5);
	}

	CPPUNIT_ASSERT(true);
}

void GraphClusteringTest::SimpleFractional() {
	gmm::row_matrix<gmm::wsvector<double> > K(4, 4);
	//  0.0  0.7 -0.6 -0.4
	//  0.7  0.0  0.3  0.4
	// -0.6  0.3  0.0 -0.1
	// -0.4  0.4 -0.1  0.0
	K(0, 1) = 0.7;
	K(0, 2) = -0.6;
	K(0, 3) = -0.4;
	K(1, 2) = 0.3;
	K(1, 3) = 0.4;
	K(2, 3) = -0.1;

	Clustering::GraphClustering gc;
	gc.Initialize(K);
	gc.Solve();

	std::cout << "solution: " << gmm::row_vector(gc.Solution()) << std::endl;
}

void GraphClusteringTest::SimpleFractional2() {
	gmm::row_matrix<gmm::wsvector<double> > K(4, 4);
	// Counterexample cycle inequalities (top one), optimal fractional
	// solution with only the cycle inequalities is -1.55, optimal integral
	// solution is -1.5.
	K(0, 1) = 0.9;
	K(0, 2) = -0.9;
	K(0, 3) = -0.7;
	K(1, 2) = 0.8;
	K(1, 3) = 0.2;
	K(2, 3) = -0.9;

	Clustering::GraphClustering gc;
	gc.Initialize(K);
	gc.Solve();

	std::cout << "solution: " << gmm::row_vector(gc.Solution()) << std::endl;
}

void GraphClusteringTest::Tables() {
	gmm::row_matrix<gmm::wsvector<double> > K(9, 9);

	// which should have the optimal partitioning: {0,1,2,3,4} {5,6,7} {8}
	// ( 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 1, 0, 1 )
	double theta = 0.0;
	K(0, 1) = 0.1 + theta;	// 0
	K(0, 3) = 0.2 + theta;	// 0
	K(0, 4) = 0.3 + theta;	// 0
	K(1, 2) = 1.0 + theta;	// 0
	K(1, 3) = 0.4 + theta;	// 0
	K(2, 3) = 0.3 + theta;	// 0 
	K(2, 8) = -0.3 + theta;	// 1
	K(3, 4) = 0.4 + theta;	// 0
	K(3, 7) = -0.8 + theta;	// 1
	K(3, 8) = 0.1 + theta;	// 1
	K(4, 5) = -0.3 + theta;	// 1
	K(4, 7) = -0.2 + theta;	// 1
	K(5, 6) = 0.7 + theta;	// 0
	K(5, 7) = 0.4 + theta;	// 0
	K(5, 8) = -0.2 + theta;	// 1
	K(6, 7) = 0.3 + theta;	// 0
	K(7, 8) = 0.1 + theta;	// 1

	Clustering::GraphClustering gc;
	gc.Initialize(K);
	gc.Solve();

	// Check solution
	const std::vector<double>& sol = gc.Solution();
	std::cout << "solution: " << gmm::row_vector(sol) << std::endl;
	// TODO
	CPPUNIT_ASSERT(true);
}


void GraphClusteringTest::RandomSmall() {
	boost::mt19937 rng;
#if 0
	boost::uniform_real<double> uni(-1.0, 1.0);
	boost::variate_generator<boost::mt19937&, boost::uniform_real<double> >
		gen(rng, uni);
#endif
	boost::uniform_int<int> uni(-10, 10);
	boost::variate_generator<boost::mt19937&, boost::uniform_int<int> >
		gen(rng, uni);

	for (unsigned int n = 0; n < 10000; ++n) {
		gmm::row_matrix<gmm::wsvector<double> > K(4, 4);
		for (unsigned int row = 0; row < 4; ++row)
			for (unsigned int col = row + 1; col < 4; ++col)
				K(col, row) = K(row, col) = 0.1 * static_cast<double>(gen());

		Clustering::GraphClustering gc;
		gc.Initialize(K);
		gc.Solve();
		const std::vector<double>& sol = gc.Solution();
		const std::vector<std::pair<unsigned int, unsigned int> >& emap =
			gc.Edgemap();

		// Assert integrality
		for (unsigned int i = 0; i < sol.size(); ++i) {
			if (sol[i] > 1e-5 && sol[i] < (1.0 - 1e-5)) {
				std::cout << "FATAL: found non-integral solution for graph:"
					<< std::endl;
				std::cout << K << std::endl;
				std::cout << "Solution: " << gmm::row_vector(sol)
					<< std::endl;
				std::cout << "Order:";
				for (unsigned int p = 0; p < sol.size(); ++p)
					std::cout << " (" << emap[p].first << "-" << emap[p].second
						<< ")";
				std::cout << std::endl;

				//exit(EXIT_FAILURE);
				std::string c;
				std::cin >> c;
			}
		}
	}
}

void GraphClusteringTest::Leaf9() {
	gmm::row_matrix<gmm::wsvector<double> > K(12, 12);

	K(0, 1) = 1.75;
	K(0, 2) = 2.125;
	K(0, 3) = -0.375;
	K(0, 4) = 2.0;
	K(0, 5) = 2.625;
	K(0, 6) = 2.375;
	K(0, 7) = -1.5;
	K(0, 8) = 1.375;
	K(0, 9) = 2.5;
	K(0, 10) = -1.625;
	K(0, 11) = 2.0;

	K(1, 2) = -2.25;
	K(1, 3) = -2.625;
	K(1, 4) = -1.875;
	K(1, 5) = -1.75;
	K(1, 6) = -2.0;
	K(1, 7) = -1.5;
	K(1, 8) = -2.375;
	K(1, 9) = -1.875;
	K(1, 10) = 0.0;
	K(1, 11) = -1.5;

	K(2, 3) = 1.25;
	K(2, 4) = 2.375;
	K(2, 5) = 1.375;
	K(2, 6) = 1.625;
	K(2, 7) = 1.0;
	K(2, 8) = 2.25;
	K(2, 9) = 1.625;
	K(2, 10) = 0.0;
	K(2, 11) = 0.0;

	K(3, 4) = -0.375;
	K(3, 5) = 0.5;
	K(3, 6) = -1.25;
	K(3, 7) = -0.375;
	K(3, 8) = 1.125;
	K(3, 9) = 0.875;
	K(3, 10) = -0.25;
	K(3, 11) = 0.0;

	K(4, 5) = 0.125;
	K(4, 6) = -1.375;
	K(4, 7) = -0.875;
	K(4, 8) = 2.375;
	K(4, 9) = 2.375;
	K(4, 10) = 0.5;
	K(4, 11) = 0.75;

	K(5, 6) = 1.25;
	K(5, 7) = 1.125;
	K(5, 8) = 0.625;
	K(5, 9) = 0.75;
	K(5, 10) = 0.75;
	K(5, 11) = -0.25;

	K(6, 7) = 0.625;
	K(6, 8) = 1.0;
	K(6, 9) = 2.125;
	K(6, 10) = 0.25;
	K(6, 11) = 1.5;

	K(7, 8) = 0.75;
	K(7, 9) = -0.625;
	K(7, 10) = -1.625;
	K(7, 11) = 1.5;

	K(8, 9) = 1.875;
	K(8, 10) = 1.0;
	K(8, 11) = 0.625;

	K(9, 10) = -0.75;
	K(9, 11) = 0.75;

	K(10, 11) = 0.875;

	Clustering::GraphClustering gc;
	gc.Initialize(K);
	gc.Solve();

	// Check solution
	const std::vector<double>& sol = gc.Solution();
	std::cout << "solution: " << gmm::row_vector(sol) << std::endl;
	// TODO
	CPPUNIT_ASSERT(true);
}

void GraphClusteringTest::ReducedCostTest() {
	OsiClpSolverInterface* si = new OsiClpSolverInterface;

	// Setup [3, -1; 2, 2] x <= [2; 2.5], 2-by-2 constraint system
	CoinPackedMatrix* matrix = new CoinPackedMatrix(false, 0, 0);
	matrix->setDimensions(0, 2);
	double varLB[] = { 0.0, 0.0, };
	double varUB[] = { 1.0, 1.0, };
	double objective[] = { -3.2, 0.0, };
	si->loadProblem(*matrix, varLB, varUB, objective, NULL, NULL);
	si->setObjSense(1);

	CoinPackedVector row1;
	row1.insert(0, 3.0);
	row1.insert(1, -1.0);
	si->addRow(row1, -si->getInfinity(), 2.0);
	CoinPackedVector row2;
	row2.insert(0, 2.0);
	row2.insert(1, 2.0);
	si->addRow(row2, -si->getInfinity(), 2.5);
	CoinPackedVector row3;
	row3.insert(0, 1.0);
	row3.insert(1, 0.0);
	si->addRow(row3, -si->getInfinity(), 0.8);


	// Solve
	si->resolve();
	CPPUNIT_ASSERT(si->isProvenOptimal());

//	CPPUNIT_ASSERT_DOUBLES_EQUAL(si->getObjValue(), -2.6, 1e-5);
	const double* primal = si->getColSolution();
	std::cout << "primal: " << primal[0] << ", " << primal[1] << std::endl;
//	CPPUNIT_ASSERT_DOUBLES_EQUAL(primal[0], 0.8125, 1e-5);
//	CPPUNIT_ASSERT_DOUBLES_EQUAL(primal[1], 0.4375, 1e-5);
	const double* dual = si->getRowPrice();
	std::cout << "dual: " << dual[0] << ", " << dual[1] << ", " << dual[2] << std::endl;
//	CPPUNIT_ASSERT_DOUBLES_EQUAL(dual[0], -0.8, 1e-5);
//	CPPUNIT_ASSERT_DOUBLES_EQUAL(dual[1], -0.4, 1e-5);

	const double* row_activity = si->getRowActivity();
	for (int n = 0 ; n < si->getNumRows(); ++n) {
		std::cout << "row " << n << ", activity " << row_activity[n] << std::endl;
	}

	// TODO: check reduced costs
	double* red_dual = new double[si->getNumRows()];
	double* red_cost = new double[si->getNumCols()];
	si->enableSimplexInterface(true);
	si->getReducedGradient(red_cost, red_dual, objective);
	const double* red_cost_2 = si->getReducedCost();
	std::cout << "Reduced costs" << std::endl;
	for (int n = 0; n < si->getNumCols(); ++n) {
		std::cout << "  variable " << n << ", reduced cost " << red_cost[n]
			<< ", red cost si " << red_cost_2[n] << std::endl;
	}
	delete[] red_dual;
	delete[] red_cost;
}

void GraphClusteringTest::TablesStability() {
	gmm::row_matrix<gmm::wsvector<double> > K(9, 9);

	// which should have the optimal partitioning: {0,1,2,3,4} {5,6,7} {8}
	// ( 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 1, 0, 1 )
	double theta = 0.0;
	K(0, 1) = 0.1 + theta;	// 0
	K(0, 3) = 0.2 + theta;	// 0
	K(0, 4) = 0.3 + theta;	// 0
	K(1, 2) = 1.0 + theta;	// 0
	K(1, 3) = 0.4 + theta;	// 0
	K(2, 3) = 0.3 + theta;	// 0 
	K(2, 8) = -0.3 + theta;	// 1
	K(3, 4) = 0.4 + theta;	// 0
	K(3, 7) = -0.8 + theta;	// 1
	K(3, 8) = 0.1 + theta;	// 1
	K(4, 5) = -0.3 + theta;	// 1
	K(4, 7) = -0.2 + theta;	// 1
	K(5, 6) = 0.7 + theta;	// 0
	K(5, 7) = 0.4 + theta;	// 0
	K(5, 8) = -0.2 + theta; // 1
	K(6, 7) = 0.3 + theta;	// 0
	K(7, 8) = 0.1 + theta;	// 1

	Clustering::GraphClustering gc;
	gc.Initialize(K);
//	gc.Solve(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
//	gc.Solve(-1, 1);
	gc.Solve(0.21, 0.21);

	const std::vector<double>& solution = gc.Solution();
	std::vector<double> d(solution.size());
	std::fill(d.begin(), d.end(), 1.0);
	double theta_min;
	double theta_max;
	gc.ComputeStabilityRangeLP(d, theta_min, theta_max);
	std::cout << "d = 1, min,max: [" << theta_min << ", " << theta_max << "]" << std::endl;

#if 0
	const std::vector<double>& solution = gc.Solution();
	std::vector<double> d(solution.size());
	std::fill(d.begin(), d.end(), 1.0);
//	d[9] = 1.0;

	std::vector<double> theta_min;
	std::vector<double> theta_max;
	gc.ComputeStabilityBox(d, theta_min, theta_max);
	for (unsigned int n = 0; n < solution.size(); ++n) {
		std::cout << "edge " << n << ", coeff " << gc.ObjectiveCoefficients()[n]
			<< ", sol " << solution[n] << ", theta in ["
			<< theta_min[n] << " ; " << theta_max[n] << "]" << std::endl;
	}

	double stable_min;
	double stable_max;
	gc.ComputeStabilityRange(d, stable_min, stable_max);
	std::cout << "Guaranteed stability within [" << stable_min
		<< " ; " << stable_max << "]" << std::endl;
#endif

	CPPUNIT_ASSERT(true);
}

void GraphClusteringTest::HeuristicKLTwo() {
	gmm::row_matrix<gmm::wsvector<double> > K(6, 6);

	K(0, 0) = K(0, 1) = K(0, 2) = 0.388889;
	K(0, 3) = K(0, 4) = -0.611111;
	K(0, 5) = 1.388889;
	K(1, 1) = K(1, 2) = 0.388889;
	K(1, 3) = K(1, 4) = K(1, 5) = -0.611111;
	K(2, 2) = 0.388889;
	K(2, 3) = K(2, 4) = K(2, 5) = -0.611111;
	K(3, 3) = K(3, 4) = K(3, 5) = 0.388889;
	K(4, 4) = K(4, 5) = 0.388889;
	K(5, 5) = 0.388889;

	Clustering::GraphClustering gc;
	gc.Initialize(K);

	__gnu_cxx::hash_set<unsigned int> part1;
	__gnu_cxx::hash_set<unsigned int> part2;
	part1.insert(0);
	part1.insert(1);
	part1.insert(2);
	part1.insert(3);
	part1.insert(4);
	part1.insert(5);
	gc.SolveHeuristicallyKLTwoPartitions(part1, part2);

	CPPUNIT_ASSERT(true);
}

void GraphClusteringTest::HeuristicKL() {
	boost::mt19937 rng;
	boost::uniform_int<int> uni(-10, 10);
	boost::variate_generator<boost::mt19937&, boost::uniform_int<int> >
		gen(rng, uni);

	unsigned int rowsize = 8;
	gmm::row_matrix<gmm::wsvector<double> > K(rowsize, rowsize);
	for (unsigned int row = 0; row < gmm::mat_nrows(K); ++row)
		for (unsigned int col = row + 1; col < gmm::mat_ncols(K); ++col)
			K(col, row) = K(row, col) = 0.1 * static_cast<double>(gen());

	Clustering::GraphClustering gc;
	gc.Initialize(K);

	__gnu_cxx::hash_set<unsigned int> part1;
	__gnu_cxx::hash_set<unsigned int> part2;
	for (unsigned int idx = 0; idx < rowsize; ++idx)
		part1.insert(idx);

	double total_delta = gc.SolveHeuristicallyKL();
	double heur_obj = gc.Objective();
	std::cout << "heur_obj = " << heur_obj << std::endl;
	gc.Solve();
	double lp_obj = gc.Objective();

	// The total delta should be exactly the objective
	std::cout << "total_delta = " << total_delta << std::endl;
	std::cout << "heur_obj = " << heur_obj << std::endl;
	std::cout << "lp_obj = " << lp_obj << std::endl;
	CPPUNIT_ASSERT_DOUBLES_EQUAL(total_delta, heur_obj, 1e-3);
	CPPUNIT_ASSERT(total_delta >= (lp_obj - 1e-8));
}

void GraphClusteringTest::HeuristicKLTwo2() {
	boost::mt19937 rng;
	boost::uniform_int<int> uni(-10, 10);
	boost::variate_generator<boost::mt19937&, boost::uniform_int<int> >
		gen(rng, uni);

	for (unsigned int tries = 0; tries < 50; ++tries) {
		unsigned int rowsize = 40;
		gmm::row_matrix<gmm::wsvector<double> > K(rowsize, rowsize);
		for (unsigned int row = 0; row < gmm::mat_nrows(K); ++row)
			for (unsigned int col = row + 1; col < gmm::mat_ncols(K); ++col)
				K(col, row) = K(row, col) = 0.1 * static_cast<double>(gen());

		Clustering::GraphClustering gc;
		gc.Initialize(K);

		__gnu_cxx::hash_set<unsigned int> part1;
		__gnu_cxx::hash_set<unsigned int> part2;
		for (unsigned int idx = 0; idx < rowsize; ++idx)
			part1.insert(idx);

		double total_delta = gc.SolveHeuristicallyKLTwoPartitions(part1, part2);
		double manual_objective = 0.0;
		for (__gnu_cxx::hash_set<unsigned int>::const_iterator pi1 = part1.begin();
			pi1 != part1.end(); ++pi1) {
			for (__gnu_cxx::hash_set<unsigned int>::const_iterator pi2 = part2.begin();
				pi2 != part2.end(); ++pi2) {
				manual_objective += K(*pi1, *pi2);
			}
		}

		// The total delta should be exactly the objective
		std::cout << "total_delta = " << total_delta << std::endl;
		std::cout << "manual_objective = " << manual_objective << std::endl;
		CPPUNIT_ASSERT_DOUBLES_EQUAL(total_delta, manual_objective, 1e-3);
	}
}

int main(int argc, char **argv) {
	CPPUNIT_NS::TestResult controller;
	CPPUNIT_NS::TestResultCollector result;
	controller.addListener(&result);        
	CPPUNIT_NS::BriefTestProgressListener progress;
	controller.addListener(&progress);      

	CPPUNIT_NS::TestRunner runner;
	runner.addTest(CPPUNIT_NS::TestFactoryRegistry::getRegistry().makeTest());
	runner.run(controller);

	CPPUNIT_NS::CompilerOutputter outputter(&result, CPPUNIT_NS::stdCOut());
	outputter.write(); 

	return (result.wasSuccessful() ? 0 : 1);
}


